#include <ctype.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "rpi_artnet.h"
#include "rpi_pixleds.h"
#include "rpi_smi_defs.h"

#include "rpi_selector.h"

/*  Command-line parameters */
bool IsTestMode = false;
int chanLedCount = 0;

/* Global */
MEM_MAP vc_mem;
TXDATA_T *txdata;

void parseCommandLineArgs(int argc, char const *argv[]);

void execute_test_mode();

void execute_artnet_mode();

void execute_autonomous_mode();

void execute_autonomous2_mode();

int main(int argc, char const *argv[]) {
  int previousMode = 0;

  // setup
  parseCommandLineArgs(argc, argv);

  signal(SIGINT, terminate);

  setup_selector();

  // setup led
  map_devices();
  init_smi(LED_NCHANS > 8 ? SMI_16_BITS : SMI_8_BITS, SMI_TIMING);
  map_uncached_mem(&vc_mem, VC_MEM_SIZE);

  setup_smi_dma(&vc_mem, TX_BUFF_LEN(chanLedCount), &txdata);

  artnet_init();

  // loop
  while (1) {
    int mode = get_selector_position();

    if (mode != previousMode) {
#if defined(_DEBUG)
      print_selector();
#endif
    }

    switch (mode) {
    case 0:
      // mode test
      execute_test_mode();
      break;
    case 1:
      // artnet mode
      execute_artnet_mode();
      break;
    case 2:
      execute_autonomous_mode();
      break;
    case 3:
      // autonomous mode 2
      execute_autonomous2_mode();
      break;

    default:
      printf("error in selector \n");
      break;
    }
    previousMode = mode;
  }

  //   printf("%s %u LED%s per channel, %u channels\n", IsTestMode ? "Testing" : "Setting",
  //   chanLedCount,
  //          chanLedCount == 1 ? "" : "s", LED_NCHANS);

  //   for (size_t colorIndex = 0; colorIndex < 3; ++colorIndex) {
  //     for (size_t i = 0; i < chanLedCount; ++i) {
  //       set_color(on_rgbs[colorIndex], &tx_buffer[LED_TX_OSET(i)]);
  //     }

  // #if LED_NCHANS <= 8
  //     swap_bytes(tx_buffer, TX_BUFF_SIZE(chanLedCount));
  // #endif

  //     memcpy(txdata, tx_buffer, TX_BUFF_SIZE(chanLedCount));
  //     start_smi(&vc_mem);
  //     usleep(CHASE_MSEC * 1000);
  //     sleep(1);
  //   }

  //   artnet_init();

  //   // loops
  //   if (IsTestMode) {
  //     while (1) {
  //       test_leds();
  //       sleep(3);
  //     }
  //   } else {
  //     while (1) {
  //       artDmx_t *artDmx;
  //       if (artnet_read(&artDmx) == OpDmx) {
  //         uint16_t dmxLength = (artDmx->lengthHi << 8) | artDmx->lengthLo;
  //         unsigned int ledCountInFrame = dmxLength / 3;
  //         uint16_t maxBound = ledCountInFrame < chanLedCount ? ledCountInFrame : chanLedCount;
  //         unsigned int universe = artDmx->subUni & (LED_NCHANS - 1);
  //         for (size_t i = 0; i < maxBound; ++i) {
  //           uint8_t *rgb = artDmx->data + (i * 3);
  //           rgb_data[i][universe] = (rgb[0] << 16) | (rgb[1] << 8) | rgb[2];
  //           rgb_txdata(rgb_data[i], &tx_buffer[LED_TX_OSET(i)]);
  //         }

  // #if LED_NCHANS <= 8
  //         swap_bytes(tx_buffer, TX_BUFF_SIZE(chanLedCount));
  // #endif

  //         memcpy(txdata, tx_buffer, TX_BUFF_SIZE(chanLedCount));
  //         // enable_dma(DMA_CHAN);
  //         start_smi(&vc_mem);
  //         // usleep(10);
  //         // while (dma_active(DMA_CHAN))
  //         //     usleep(10);
  //       }
  //     }
  //   }
  //   terminate(0);
}

void parseCommandLineArgs(int argc, char const *argv[]) {
  int args = 0;

  while (argc > ++args) // Process command-line args
  {
    if (argv[args][0] == '-') {
      switch (toupper(argv[args][1])) {
      case 'N': // -N: number of LEDs per channel
        if (args >= argc - 1)
          fprintf(stderr, "Error: no numeric value\n");
        else
          chanLedCount = atoi(argv[++args]);
        break;
      case 'T': // -T: test mode
        IsTestMode = true;
        break;
      default: // Otherwise error
        printf("Unrecognised option '%c'\n", argv[args][1]);
        printf("Options:\n"
               "  -n num    number of LEDs per channel\n"
               "  -t        Test mode (flash LEDs)\n");
        exit(1);
      }
    }
  }
}
// Pointer to uncached Tx data buffer
TXDATA_T tx_buffer[TX_BUFF_LEN(CHAN_MAXLEDS)]; // Tx buffer for assembling data

void execute_test_mode() {
  // RGB values for test mode (1 value for each of 16 channels)
  uint32_t on_rgbs[] = {0xef0000, 0x00ef00, 0x0000ef, 0xefef00, 0xef00ef, 0x00efef, 0xefefef};
  uint32_t off_rgbs = 0x000000;

  static int i = 0, offset = 0, ledIndex = 0;

  if (ledIndex < chanLedCount) {
    set_color(ledIndex <= offset % chanLedCount ? on_rgbs[i] : off_rgbs,
              &tx_buffer[LED_TX_OSET(ledIndex)]);
    ++ledIndex;

  } else {
    ledIndex = 0;

#if LED_NCHANS <= 8
    swap_bytes(tx_buffer, TX_BUFF_SIZE(chanLedCount));
#endif

    memcpy(txdata, tx_buffer, TX_BUFF_SIZE(chanLedCount));
    start_smi(&vc_mem);
    usleep(CHASE_MSEC * 1000);
    if (offset < chanLedCount) {
      ++offset;
    } else {
      offset = 0;
      if (i < 7) {
        ++i;
      } else {
        i = 0;
      }
    }
  }
}

void execute_artnet_mode() {
  artDmx_t *artDmx;
  // RGB data
  int rgb_data[CHAN_MAXLEDS][LED_NCHANS];

  if (artnet_read(&artDmx) == OpDmx) {
    uint16_t dmxLength = (artDmx->lengthHi << 8) | artDmx->lengthLo;
    unsigned int ledCountInFrame = dmxLength / 3;
    uint16_t maxBound = ledCountInFrame < chanLedCount ? ledCountInFrame : chanLedCount;
    unsigned int universe = artDmx->subUni & (LED_NCHANS - 1);
    for (size_t i = 0; i < maxBound; ++i) {
      uint8_t *rgb = artDmx->data + (i * 3);
      rgb_data[i][universe] = (rgb[0] << 16) | (rgb[1] << 8) | rgb[2];
      rgb_txdata(rgb_data[i], &tx_buffer[LED_TX_OSET(i)]);
    }

#if LED_NCHANS <= 8
    swap_bytes(tx_buffer, TX_BUFF_SIZE(chanLedCount));
#endif

    memcpy(txdata, tx_buffer, TX_BUFF_SIZE(chanLedCount));
    enable_dma(DMA_CHAN);
    start_smi(&vc_mem);
    usleep(10);
    while (dma_active(DMA_CHAN))
      usleep(10);
  }
}

void execute_autonomous_mode() {}

void execute_autonomous2_mode() {}